from __future__ import annotations

import numpy as np
from scipy.special import gammaln

from ... import marlib_cf1 as _cf1
from ..data import NHPPData

__all__ = [
    "dcf1",
    "pcf1",
    "qcf1",
    "rcf1",
    "cf1_params_power",
    "cf1_params_linear",
    "cf1_params_init",
    "cf1llf",
]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _restore_order(values: np.ndarray, order: np.ndarray) -> np.ndarray:
    out = np.empty_like(values)
    out[order] = values
    return out


def dcf1(x, alpha=1.0, rate=1.0, log: bool = False, eps: float = 1.0e-8, unif_factor: float = 1.01):
    x_arr = np.atleast_1d(_as_float_array(x))
    if x_arr.size == 0:
        return x_arr
    s = np.argsort(x_arr)
    dx = np.diff(np.concatenate([[0.0], x_arr[s]]))
    res = _cf1.cf1pdf(dx, _as_float_array(alpha), _as_float_array(rate), eps, unif_factor, log)
    out = _restore_order(np.asarray(res, dtype=float), s)
    if np.ndim(x) == 0:
        return float(out[0])
    return out


def pcf1(
    q,
    alpha=1.0,
    rate=1.0,
    lower_tail: bool = True,
    log_p: bool = False,
    eps: float = 1.0e-8,
    unif_factor: float = 1.01,
):
    q_arr = np.atleast_1d(_as_float_array(q))
    if q_arr.size == 0:
        return q_arr
    s = np.argsort(q_arr)
    dx = np.diff(np.concatenate([[0.0], q_arr[s]]))
    res = _cf1.cf1cdf(dx, _as_float_array(alpha), _as_float_array(rate), eps, unif_factor, lower_tail, log_p)
    out = _restore_order(np.asarray(res, dtype=float), s)
    if np.ndim(q) == 0:
        return float(out[0])
    return out


def qcf1(p, alpha=1.0, rate=1.0, lower_tail: bool = True, log_p: bool = False):
    raise NotImplementedError("qcf1 is not implemented")


def rcf1(n: int, alpha=1.0, rate=1.0, scramble: bool = True):
    if n < 0:
        raise ValueError("n must be >= 0")
    res = _cf1.cf1sample(int(n), _as_float_array(alpha), _as_float_array(rate))
    res = np.asarray(res, dtype=float)
    if scramble:
        rng = np.random.default_rng()
        return rng.permutation(res)
    return res


def cf1_params_power(n: int, scale: float, shape: float) -> dict:
    n = int(n)
    rate = np.zeros(n, dtype=float)
    p = np.exp(1.0 / (n - 1) * np.log(shape))
    total = 1.0
    tmp = 1.0
    for i in range(1, n):
        tmp = tmp * (i + 1) / (i * p)
        total += tmp
    base = total / (n * scale)
    tmp = base
    for i in range(n):
        rate[i] = tmp
        tmp *= p
    return {"alpha": np.full(n, 1.0 / n, dtype=float), "rate": rate}


def cf1_params_linear(n: int, scale: float, shape: float) -> dict:
    n = int(n)
    rate = np.zeros(n, dtype=float)
    al = (shape - 1.0) / (n - 1)
    total = 1.0
    for i in range(1, n + 1):
        total += (i + 1) / (al * i + 1.0)
    base = total / (n * scale)
    for i in range(1, n + 1):
        rate[i - 1] = base * (al + i + 1.0)
    return {"alpha": np.full(n, 1.0 / n, dtype=float), "rate": rate}


def cf1_params_init(
    n: int,
    data: NHPPData,
    shape_init=(1, 4, 16, 64, 256, 1024),
    scale_init=(0.5, 1.0, 2.0),
    maxiter_init: int = 5,
    verbose: bool = False,
) -> dict | None:
    if verbose:
        print("Initializing CF1 ...")

    m = float(data.mean)
    max_llf = -np.inf
    best = None

    core_data = {
        "time": np.asarray(data.time, dtype=float),
        "fault": np.asarray(data.fault, dtype=np.int64),
        "type": np.asarray(data.type, dtype=np.int64),
    }

    for fn in (cf1_params_power, cf1_params_linear):
        for x in scale_init:
            for s in shape_init:
                param = {"omega": float(data.total), **fn(n, scale=m * float(x), shape=float(s))}
                try:
                    res = {"param": param}
                    for _ in range(int(maxiter_init)):
                        res = _cf1.em_cf1_emstep(res["param"], core_data)
                    llf = float(res["llf"])
                except Exception:
                    if verbose:
                        print("-", end="")
                    continue

                if np.isfinite(llf):
                    if llf > max_llf:
                        max_llf = llf
                        best = res["param"]
                        if verbose:
                            print("o", end="")
                    else:
                        if verbose:
                            print("x", end="")
                else:
                    if verbose:
                        print("-", end="")
        if verbose:
            print()

    return best


# def cf1llf(data: NHPPData, omega: float, alpha, rate) -> float:
#     omega = float(omega)
#     alpha = _as_float_array(alpha)
#     rate = _as_float_array(rate)

#     te = float(np.sum(data.time))
#     llf = -omega * float(pcf1(te, alpha=alpha, rate=rate, lower_tail=True))

#     tt = np.asarray(data.time, dtype=float)[np.asarray(data.type, dtype=int) == 1]
#     if tt.size != 0:
#         llf += np.sum(np.log(omega * dcf1(np.cumsum(tt), alpha=alpha, rate=rate)))

#     type_mask = np.asarray(data.type, dtype=int) == 0
#     gt = np.asarray(data.time, dtype=float)[type_mask]
#     if gt.size != 0:
#         barp = -np.diff(
#             pcf1(
#                 np.concatenate([[0.0], np.cumsum(gt)]),
#                 alpha=alpha,
#                 rate=rate,
#                 lower_tail=False,
#             )
#         )
#         faults0 = np.asarray(data.fault, dtype=float)[type_mask]
#         i = faults0 != 0
#         if np.any(i):
#             llf += np.sum(faults0[i] * np.log(omega * barp[i]) - gammaln(faults0[i] + 1.0))

#     return float(llf)

def cf1llf(data: NHPPData, omega: float, alpha, rate) -> float:
    omega = float(omega)
    if not np.isfinite(omega) or omega <= 0.0:
        return -np.inf

    alpha = _as_float_array(alpha)
    rate = _as_float_array(rate)

    te = float(np.sum(data.time))
    llf = -omega * float(pcf1(te, alpha=alpha, rate=rate, lower_tail=True))

    # type==1 contribution (event-time/continuous part)
    tt = np.asarray(data.time, dtype=float)[np.asarray(data.type, dtype=int) == 1]
    if tt.size != 0:
        dens = np.asarray(dcf1(np.cumsum(tt), alpha=alpha, rate=rate), dtype=float)
        x = omega * dens
        if np.any(x <= 0) or not np.all(np.isfinite(x)):
            return -np.inf
        llf += float(np.sum(np.log(x)))

    # type==0 contribution (grouped part)
    type_mask = np.asarray(data.type, dtype=int) == 0
    gt = np.asarray(data.time, dtype=float)[type_mask]
    if gt.size != 0:
        tgrid = np.concatenate([[0.0], np.cumsum(gt)])
        S = np.asarray(pcf1(tgrid, alpha=alpha, rate=rate, lower_tail=False), dtype=float)

        # Monotonicity correction (optional but useful)
        S = np.minimum.accumulate(S)

        barp = -np.diff(S)
        faults0 = np.asarray(data.fault, dtype=float)[type_mask]
        idx = faults0 != 0
        if np.any(idx):
            x = omega * barp[idx]
            if np.any(x <= 0) or not np.all(np.isfinite(x)):
                return -np.inf
            llf += float(np.sum(faults0[idx] * np.log(x) - gammaln(faults0[idx] + 1.0)))

    return float(llf)


from __future__ import annotations

import numpy as np

from ..base import NHPPModel
from ..data import NHPPData
from ..dists.pareto2 import dpareto2, ppareto2
from ... import _core


class Pareto2NHPP(NHPPModel):
    name = "pareto"
    df = 3

    def __init__(self, *, omega0: float = 1.0, shape0: float = 1.0, scale0: float = 1.0):
        super().__init__(omega0=float(omega0), shape0=float(shape0), scale0=float(scale0))
        self._set_fitted_params(np.array([omega0, shape0, scale0], dtype=float))

    def param_names(self) -> list[str]:
        return ["omega", "shape", "scale"]

    def init_params(self, data: NHPPData) -> np.ndarray:
        return np.array([1.0, 1.0, 1.0], dtype=float)

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        return _core.em_pareto_emstep(np.asarray(params, dtype=float), data.to_core_dict())

    def omega_(self) -> float:
        return float(self.params_[0])

    def shape_(self) -> float:
        return float(self.params_[1])

    def scale_(self) -> float:
        return float(self.params_[2])

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * ppareto2(t, shape=self.shape_(), scale=self.scale_(), lower_tail=True)

    def dmvf(self, t):
        t = np.asarray(t, dtype=float)
        t2 = np.concatenate([[0.0], t])
        return np.diff(self.mvf(t2))

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dpareto2(t, shape=self.shape_(), scale=self.scale_(), log=False)





from __future__ import annotations

import numpy as np
import matplotlib.pyplot as plt

from .data import NHPPData
from .base import NHPPModel


def _normalize_results(results):
    if isinstance(results, NHPPModel):
        return {results.name: results}
    if isinstance(results, dict):
        return results
    if isinstance(results, (list, tuple)):
        out = {}
        for item in results:
            if isinstance(item, NHPPModel):
                out[item.name] = item
            else:
                raise TypeError("Unsupported results type")
        return out
    raise TypeError("Unsupported results type")


def plot_mvf(
    data: NHPPData,
    results,
    *,
    xlab: str = "time",
    ylab: str = "# of faults",
    datalab: str = "data",
    xmax=None,
    ymax=None,
    num_points: int = 300,
    ax=None,
):
    results = _normalize_results(results)
    present = float(np.sum(data.time))
    if xmax is None:
        xmax = present * 1.2

    n = data.fault + data.type
    x_data = np.cumsum(data.time)[n != 0]
    y_data = np.cumsum(n)[n != 0]

    if ax is None:
        _, ax = plt.subplots()

    ax.scatter(x_data, y_data, label=datalab)
    ax.axvline(present, linestyle="dotted")

    tgrid = np.linspace(0.0, float(xmax), num_points)
    for name, res in results.items():
        ax.plot(tgrid, res.mvf(tgrid), label=name)

    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    ax.set_xlim(0.0, float(xmax))
    if ymax is not None:
        ax.set_ylim(0.0, float(ymax))

    ax.legend()
    return ax


def plot_dmvf(
    data: NHPPData,
    results,
    *,
    xlab: str = "time",
    ylab: str = "# of faults",
    datalab: str = "data",
    xmax=None,
    ymax=None,
    ax=None,
):
    results = _normalize_results(results)
    t = np.cumsum(data.time).astype(float)
    n = (data.fault + data.type).astype(float)

    if ax is None:
        _, ax = plt.subplots()

    ax.bar(t, n, width=np.maximum(data.time, 1e-12), alpha=0.5, label=datalab, align="center")

    for name, res in results.items():
        ax.plot(t, res.dmvf(t), label=name)

    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    if xmax is not None:
        ax.set_xlim(0.0, float(xmax))
    if ymax is not None:
        ax.set_ylim(0.0, float(ymax))

    ax.legend()
    return ax


def plot_rate(
    data: NHPPData,
    results,
    *,
    xlab: str = "time",
    ylab: str = "detection rate",
    datalab: str = "data",
    xmax=None,
    ymax=None,
    num_points: int = 300,
    ax=None,
):
    results = _normalize_results(results)
    present = float(np.sum(data.time))
    if xmax is None:
        xmax = present * 1.2

    t = np.cumsum(data.time)
    n = data.fault + data.type
    rate_emp = np.divide(n, data.time, where=data.time > 0)

    if ax is None:
        _, ax = plt.subplots()

    ax.bar(t, rate_emp, width=np.maximum(data.time, 1e-12), alpha=0.3, label=datalab)
    ax.axvline(present, linestyle="dotted")

    tgrid = np.linspace(0.1, float(xmax), num_points)
    for name, res in results.items():
        ax.plot(tgrid, res.rate(tgrid), label=name)

    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    ax.set_xlim(0.1, float(xmax))
    if ymax is not None:
        ax.set_ylim(0.0, float(ymax))

    ax.legend()
    return ax



from __future__ import annotations

import numpy as np
from scipy.stats import truncnorm

__all__ = ["dtnorm", "ptnorm", "qtnorm", "rtnorm"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _truncnorm_dist(mean, sd):
    a = (0.0 - mean) / sd
    b = np.inf
    return truncnorm(a=a, b=b, loc=mean, scale=sd)


# -------------------------
# Truncated normal at 0 (left-truncated)
# -------------------------

def dtnorm(x, mean=0.0, sd=1.0, log: bool = False):
    """
    Density of Normal(mean, sd) truncated to [0, +inf).
    """
    x = _as_float_array(x)
    mean = float(mean)
    sd = float(sd)
    if sd <= 0:
        raise ValueError("sd must be > 0.")

    dist = _truncnorm_dist(mean, sd)
    if log:
        return dist.logpdf(x)
    return dist.pdf(x)


def ptnorm(q, mean=0.0, sd=1.0, lower_tail: bool = True, log_p: bool = False):
    """
    CDF / upper-tail probability for Normal(mean, sd) truncated to [0, +inf).
    """
    q = _as_float_array(q)
    mean = float(mean)
    sd = float(sd)
    if sd <= 0:
        raise ValueError("sd must be > 0.")

    dist = _truncnorm_dist(mean, sd)
    if lower_tail:
        v = dist.cdf(q)
    else:
        v = dist.sf(q)
    if log_p:
        return np.log(v)
    return v


def qtnorm(p, mean=0.0, sd=1.0, lower_tail: bool = True, log_p: bool = False):
    """
    Quantile for Normal(mean, sd) truncated to [0, +inf).
    """
    p = _as_float_array(p)
    mean = float(mean)
    sd = float(sd)
    if sd <= 0:
        raise ValueError("sd must be > 0.")

    if log_p:
        p = np.exp(p)

    if not lower_tail:
        p = 1.0 - p

    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")

    dist = _truncnorm_dist(mean, sd)
    if lower_tail:
        return dist.ppf(p)
    return dist.isf(p)


def rtnorm(n: int, mean=0.0, sd=1.0, rng: np.random.Generator | None = None):
    """
    Random variates from Normal(mean, sd) truncated to [0, +inf).
    """
    if n < 0:
        raise ValueError("n must be >= 0.")
    rng = rng if rng is not None else np.random.default_rng()
    dist = _truncnorm_dist(float(mean), float(sd))
    return dist.rvs(size=int(n), random_state=rng)



from __future__ import annotations

from .data import NHPPData
from .base import NHPPModel


def compare(models: list[NHPPModel], data: NHPPData, *, criterion: str = "AIC"):
    fitted = [m.fit(data) for m in models]
    crit = criterion.upper()
    if crit != "AIC":
        raise ValueError(f"Unsupported criterion: {criterion}")
    best = min(fitted, key=lambda m: m.aic_)
    return fitted, best



from __future__ import annotations

from abc import ABC, abstractmethod
import numpy as np
from .data import NHPPData


class NHPPModel(ABC):
    name: str
    df: int

    def __init__(self, **hyperparams):
        self._hyperparams = dict(hyperparams)
        self._fitted = False
        self.params_: np.ndarray | None = None
        self.llf_: float | None = None
        self.aic_: float | None = None
        self.converged_: bool | None = None
        self.n_iter_: int | None = None
        self.initial_params_: np.ndarray | None = None
        self.aerror_: float | None = None
        self.rerror_: float | None = None
        self.data_: NHPPData | None = None

    def get_params(self, deep: bool = True) -> dict:
        return dict(self._hyperparams)

    def set_params(self, **params):
        for k, v in params.items():
            if k not in self._hyperparams:
                raise ValueError(f"Invalid parameter {k} for estimator {self.__class__.__name__}.")
            self._hyperparams[k] = v
        return self

    def _get_fitted_params(self) -> np.ndarray:
        if self.params_ is None:
            raise RuntimeError("Fitted parameters are not set.")
        return self.params_

    def _set_fitted_params(self, params: np.ndarray):
        self.params_ = np.asarray(params, dtype=float)

    def set_data(self, data: NHPPData):
        self.data_ = data

    @property
    def params(self) -> np.ndarray:
        if not self._fitted:
            raise RuntimeError("Model is not fitted.")
        return self.params_

    @property
    def params_dict_(self) -> dict:
        if not self._fitted or self.params_ is None:
            raise RuntimeError("Model is not fitted.")
        names = self.param_names()
        vals = np.asarray(self.params_, dtype=float)
        if len(names) != len(vals):
            names = [f"p{i}" for i in range(len(vals))]
        return {k: float(v) for k, v in zip(names, vals)}

    def fit(
        self,
        data: NHPPData,
        *,
        initialize: bool = True,
        maxiter: int = 2000,
        reltol: float = 1.0e-6,
        abstol: float = 1.0e-3,
        trace: bool = False,
        verbose: bool = False,
        printsteps: int = 50,
        **kwargs,
    ):
        from ._em import emfit_internal

        r = emfit_internal(
            self,
            data,
            initialize=initialize,
            maxiter=maxiter,
            reltol=reltol,
            abstol=abstol,
            trace=trace,
            verbose=verbose,
            printsteps=printsteps,
            **kwargs,
        )

        self._fitted = True
        self.initial_params_ = r.initial
        self.llf_ = r.llf
        self.converged_ = r.convergence
        self.n_iter_ = r.iter
        self.aerror_ = r.aerror
        self.rerror_ = r.rerror
        self.aic_ = float(-2.0 * self.llf_ + 2.0 * self.df)
        return self

    @abstractmethod
    def param_names(self) -> list[str]:
        ...

    @abstractmethod
    def init_params(self, data: NHPPData) -> np.ndarray:
        ...

    @abstractmethod
    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        ...

    @abstractmethod
    def mvf(self, t):
        ...

    @abstractmethod
    def intensity(self, t):
        ...

    def dmvf(self, t):
        return self.intensity(t)

    def rate(self, t):
        return self.intensity(t)



from __future__ import annotations

import numpy as np

from ..base import NHPPModel
from ..data import NHPPData
from ... import marlib_cf1 as _cf1
from ..dists.cf1 import dcf1, pcf1, cf1_params_init, cf1llf


class CanonicalPhaseTypeNHPP(NHPPModel):
    name = "cf1"

    def __init__(
        self,
        n: int,
        *,
        omega0: float = 1.0,
        alpha0: np.ndarray | None = None,
        rate0: np.ndarray | None = None,
    ):
        if n <= 0:
            raise ValueError("n must be positive")
        if alpha0 is None:
            alpha0 = np.full(n, 1.0 / n, dtype=float)
        if rate0 is None:
            rate0 = np.ones(n, dtype=float)
        alpha0 = np.asarray(alpha0, dtype=float)
        rate0 = np.asarray(rate0, dtype=float)
        if alpha0.shape[0] != n or rate0.shape[0] != n:
            raise ValueError("alpha0 and rate0 must have length n")

        self.n = int(n)
        self.df = int(2 * n)
        self.name = f"cph{n}"

        super().__init__(n=int(n), omega0=float(omega0), alpha0=alpha0, rate0=rate0)
        self._set_fitted_params(
            np.concatenate([[float(omega0)], alpha0.astype(float), rate0.astype(float)])
        )

    def param_names(self) -> list[str]:
        alpha_names = [f"alpha{i + 1}" for i in range(self.n)]
        rate_names = [f"rate{i + 1}" for i in range(self.n)]
        return ["omega", *alpha_names, *rate_names]

    def init_params(self, data: NHPPData) -> np.ndarray:
        params = cf1_params_init(self.n, data)
        if params is None:
            omega = float(data.total)
            alpha = np.full(self.n, 1.0 / self.n, dtype=float)
            if data.mean > 0:
                rate = np.full(self.n, 1.0 / data.mean, dtype=float)
            else:
                rate = np.ones(self.n, dtype=float)
        else:
            omega = float(params["omega"])
            alpha = np.asarray(params["alpha"], dtype=float)
            rate = np.asarray(params["rate"], dtype=float)
        return np.concatenate([[omega], alpha, rate])

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        params = np.asarray(params, dtype=float)
        if params.size != 1 + 2 * self.n:
            raise ValueError("invalid parameter length")

        omega = float(params[0])
        alpha = np.asarray(params[1 : 1 + self.n], dtype=float)
        rate = np.asarray(params[1 + self.n : 1 + 2 * self.n], dtype=float)

        pdata = {
            "time": np.asarray(data.time, dtype=float),
            "fault": np.asarray(data.fault, dtype=np.int64),
            "type": np.asarray(data.type, dtype=np.int64),
        }
        pparams = {
            "omega": omega,
            "alpha": alpha,
            "rate": rate,
        }

        res = _cf1.em_cf1_emstep(
            pparams,
            pdata,
            kwargs.get("eps", 1.0e-8),
            kwargs.get("ufactor", 1.01),
        )

        new_omega = float(res["param"]["omega"])
        new_alpha = np.asarray(res["param"]["alpha"], dtype=float)
        new_rate = np.asarray(res["param"]["rate"], dtype=float)

        param = np.concatenate([[new_omega], new_alpha, new_rate])

        if "pdiff" in res:
            pd_omega = float(res["pdiff"]["omega"])
            pd_alpha = np.asarray(res["pdiff"]["alpha"], dtype=float)
            pd_rate = np.asarray(res["pdiff"]["rate"], dtype=float)
            pdiff = np.concatenate([[pd_omega], pd_alpha, pd_rate])
        else:
            pdiff = param - params

        return {
            "param": param,
            "pdiff": pdiff,
            "llf": float(res["llf"]),
            "total": float(res["total"]),
        }

    def omega_(self) -> float:
        return float(self.params_[0])

    def alpha_(self) -> np.ndarray:
        return np.asarray(self.params_[1 : 1 + self.n], dtype=float)

    def rate_(self) -> np.ndarray:
        return np.asarray(self.params_[1 + self.n : 1 + 2 * self.n], dtype=float)

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * pcf1(t, alpha=self.alpha_(), rate=self.rate_(), lower_tail=True, log_p=False)

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dcf1(t, alpha=self.alpha_(), rate=self.rate_(), log=False)

    def llf(self, data: NHPPData) -> float:
        return cf1llf(data, self.omega_(), self.alpha_(), self.rate_())



"""Distribution helpers (R-like naming)."""
from __future__ import annotations

from .exp import dexp, pexp, qexp, rexp
from .gamma import dgamma, pgamma, qgamma, rgamma
from .gumbel import dgumbel, pgumbel, qgumbel, rgumbel
from .lgumbel import dlgumbel, plgumbel, qlgumbel, rlgumbel
from .llogis import dllogis, pllogis, qllogis, rllogis
from .lnorm import dlnorm, plnorm, qlnorm, rlnorm
from .pareto2 import dpareto2, ppareto2, qpareto2, rpareto2
from .cf1 import dcf1, pcf1, qcf1, rcf1
from .tgumbel import dtgumbel, ptgumbel, qtgumbel, rtgumbel
from .tlogis import dtlogis, ptlogis, qtlogis, rtlogis
from .tnorm import dtnorm, ptnorm, qtnorm, rtnorm

__all__ = [
    "dexp",
    "pexp",
    "qexp",
    "rexp",
    "dgamma",
    "pgamma",
    "qgamma",
    "rgamma",
    "dgumbel",
    "pgumbel",
    "qgumbel",
    "rgumbel",
    "dlgumbel",
    "plgumbel",
    "qlgumbel",
    "rlgumbel",
    "dllogis",
    "pllogis",
    "qllogis",
    "rllogis",
    "dlnorm",
    "plnorm",
    "qlnorm",
    "rlnorm",
    "dpareto2",
    "ppareto2",
    "qpareto2",
    "rpareto2",
    "dcf1",
    "pcf1",
    "qcf1",
    "rcf1",
    "dtgumbel",
    "ptgumbel",
    "qtgumbel",
    "rtgumbel",
    "dtlogis",
    "ptlogis",
    "qtlogis",
    "rtlogis",
    "dtnorm",
    "ptnorm",
    "qtnorm",
    "rtnorm",
]



from __future__ import annotations

import numpy as np

from ..base import NHPPModel
from ..data import NHPPData
from ..dists.tnorm import ptnorm, dtnorm
from ... import _core


class TruncatedNormalNHPP(NHPPModel):
    name = "tnorm"
    df = 3

    def __init__(self, *, omega0: float = 1.0, mu0: float = 0.0, sig0: float = 1.0):
        super().__init__(omega0=float(omega0), mu0=float(mu0), sig0=float(sig0))
        self._set_fitted_params(np.array([omega0, mu0, sig0], dtype=float))

    def param_names(self) -> list[str]:
        return ["omega", "mu", "sig"]

    def init_params(self, data: NHPPData) -> np.ndarray:
        return np.array([1.0, 0.0, data.mean], dtype=float)

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        return _core.em_tnorm_emstep(np.asarray(params, dtype=float), data.to_core_dict())

    def omega_(self) -> float:
        return float(self.params_[0])

    def mu_(self) -> float:
        return float(self.params_[1])

    def sig_(self) -> float:
        return float(self.params_[2])

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * ptnorm(t, mean=self.mu_(), sd=self.sig_(), lower_tail=True)

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dtnorm(t, mean=self.mu_(), sd=self.sig_(), log=False)





"""Gamma distribution helpers."""
from __future__ import annotations

import numpy as np
from scipy.stats import gamma as _gamma

__all__ = ["dgamma", "pgamma", "qgamma", "rgamma"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _dist(shape: float, rate: float):
    return _gamma(a=shape, scale=1.0 / rate)


def dgamma(x, shape: float = 1.0, rate: float = 1.0, log: bool = False):
    """Density of Gamma(shape, rate)."""
    x = _as_float_array(x)
    shape = float(shape)
    rate = float(rate)
    if shape <= 0:
        raise ValueError("shape must be > 0.")
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    dist = _dist(shape, rate)
    if log:
        out = dist.logpdf(x)
        return np.where(x <= 0.0, -np.inf, out)
    out = dist.pdf(x)
    return np.where(x <= 0.0, 0.0, out)


def pgamma(q, shape: float = 1.0, rate: float = 1.0, lower_tail: bool = True, log_p: bool = False):
    """CDF / upper-tail probability for Gamma(shape, rate)."""
    q = _as_float_array(q)
    shape = float(shape)
    rate = float(rate)
    if shape <= 0:
        raise ValueError("shape must be > 0.")
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    dist = _dist(shape, rate)
    if lower_tail:
        v = dist.cdf(q)
        v = np.where(q <= 0.0, 0.0, v)
    else:
        v = dist.sf(q)
        v = np.where(q <= 0.0, 1.0, v)
    if log_p:
        return np.log(v)
    return v


def qgamma(p, shape: float = 1.0, rate: float = 1.0, lower_tail: bool = True, log_p: bool = False):
    """Quantile for Gamma(shape, rate)."""
    p = _as_float_array(p)
    shape = float(shape)
    rate = float(rate)
    if shape <= 0:
        raise ValueError("shape must be > 0.")
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    if log_p:
        p = np.exp(p)
    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")
    dist = _dist(shape, rate)
    if lower_tail:
        return dist.ppf(p)
    return dist.isf(p)


def rgamma(n: int, shape: float = 1.0, rate: float = 1.0, rng: np.random.Generator | None = None):
    """Random variates from Gamma(shape, rate)."""
    if n < 0:
        raise ValueError("n must be >= 0.")
    shape = float(shape)
    rate = float(rate)
    if shape <= 0:
        raise ValueError("shape must be > 0.")
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    rng = rng if rng is not None else np.random.default_rng()
    dist = _dist(shape, rate)
    return dist.rvs(size=int(n), random_state=rng)



from __future__ import annotations

from dataclasses import dataclass
import warnings
import numpy as np

try:
    from tqdm.auto import tqdm as _tqdm
except Exception:  # pragma: no cover - optional dependency
    _tqdm = None

from .data import NHPPData
from .base import NHPPModel

__all__ = []


@dataclass
class EmFitResult:
    initial: np.ndarray
    llf: float
    convergence: bool
    iter: int
    aerror: float
    rerror: float


def _comp_error_llf(res0: dict, res1: dict) -> np.ndarray:
    sdiff = float(res1["llf"] - res0["llf"])
    aerror = abs(sdiff)
    denom = abs(float(res0["llf"]))
    rerror = aerror / denom if denom > 0 else np.inf
    return np.array([aerror, rerror, sdiff], dtype=float)


def emfit_internal(
    model: NHPPModel,
    data: NHPPData,
    *,
    initialize: bool = True,
    maxiter: int = 2000,
    reltol: float = 1.0e-6,
    abstol: float = 1.0e-3,
    trace: bool = False,
    verbose: bool = False,
    printsteps: int = 50,
    debug_dir: str | None = None,
    strict: bool = False,
    **kwargs,
) -> EmFitResult:
    import json
    import os
    if initialize:
        init = model.init_params(data)
        model._set_fitted_params(np.asarray(init, dtype=float))

    iter_ = 1
    conv = False
    initial_param = np.asarray(model._get_fitted_params(), dtype=float).copy()

    res0 = {"param": initial_param, "llf": -np.inf}
    res1 = res0
    error = np.array([np.inf, np.inf, np.inf], dtype=float)

    def _dump_state(tag: str, param, llf_val):
        if debug_dir is None:
            return
        os.makedirs(debug_dir, exist_ok=True)
        path = os.path.join(debug_dir, f"{model.name}_{tag}.json")
        payload = {
            "tag": tag,
            "param": np.asarray(param, dtype=float).tolist(),
            "llf": float(llf_val),
            "initial_param": initial_param.tolist(),
        }
        with open(path, "w", encoding="utf-8") as f:
            json.dump(payload, f, indent=2)

    pbar = None
    if verbose:
        if _tqdm is None:
            warnings.warn("tqdm is not installed; set verbose=False or install tqdm to see a progress bar.")
        else:
            pbar = _tqdm(total=maxiter, desc=f"{model.name} fit", leave=False)

    while True:
        try:
            step = model.em_step(res0["param"], data, **kwargs)
        except Exception as exc:
            _dump_state(f"exception_iter{iter_}", res0["param"], res0["llf"])
            if strict:
                raise
            warnings.warn(f"Exception in EM step: {exc}")
            break

        new_param = np.asarray(step["param"], dtype=float)
        new_llf = float(step["llf"])

        res1 = {
            "param": new_param,
            "llf": new_llf,
            "pdiff": np.asarray(step.get("pdiff", new_param - res0["param"]), dtype=float),
        }

        if not np.all(np.isfinite(new_param)):
            _dump_state(f"param_nan_iter{iter_}", new_param, new_llf)
            if strict:
                raise FloatingPointError("Non-finite parameter in EM")
            warnings.warn(f"Non-finite parameter detected at iter {iter_}")
            break

        if not np.isfinite(new_llf):
            _dump_state(f"llf_nan_iter{iter_}", new_param, new_llf)
            if strict:
                raise FloatingPointError("LLF became NaN/Inf")
            warnings.warn(f"LLF becomes NaN/Inf at iter {iter_}")
            break

        error = _comp_error_llf(res0, res1)

        if pbar is not None:
            pbar.update(1)

        if trace and (iter_ % printsteps == 0):
            print(f"llf={res1['llf']} ({error[2]:.6e}) params=({res1['param']})")

        if error[2] < 0:
            warnings.warn(f"LLF decreases: {model.name} {iter_} {error[2]:.6e}")

        if (error[0] < abstol) and (error[1] < reltol):
            conv = True
            model._set_fitted_params(res1["param"])
            break

        if iter_ >= maxiter:
            warnings.warn("Did not converge to MLE by max iteration.")
            model._set_fitted_params(res1["param"])
            break

        iter_ += 1
        res0 = res1

    if pbar is not None:
        pbar.n = min(iter_, maxiter)
        pbar.refresh()
        pbar.close()

    model.set_data(data)
    return EmFitResult(
        initial=initial_param,
        llf=float(res1["llf"]),
        convergence=bool(conv),
        iter=int(iter_),
        aerror=float(error[0]),
        rerror=float(error[1]),
    )



import numpy as np
import pytest

from pysrat import NHPPData
from pysrat.nhpp.bootstrap import eic_sample, eic
from pysrat.nhpp.models.cf1 import CanonicalPhaseTypeNHPP

pytest.importorskip("pysrat.marlib_cf1")


def _time_data():
    times = np.linspace(0.2, 4.0, 20)
    return NHPPData.from_fault_times(times=times, te=5.0)


def _group_data():
    return NHPPData.from_counts(fault=np.array([0, 1, 0, 2], dtype=int))


def test_eic_time_sample_and_summary():
    data = _time_data()
    model = CanonicalPhaseTypeNHPP(3)
    model.fit(data, initialize=True, maxiter=50)

    rng = np.random.default_rng(0)
    sample = eic_sample(model, bsample=3, initialize=True, rng=rng)
    assert sample.shape == (3,)
    assert np.all(np.isfinite(sample))

    summary = eic(model, bsample=3, alpha=0.9, initialize=True, rng=rng)
    assert set(summary.keys()) == {
        "bias",
        "bias.lower",
        "bias.upper",
        "eic",
        "eic.lower",
        "eic.upper",
    }
    assert np.isfinite(summary["eic"])


def test_eic_group_sample():
    data = _group_data()
    model = CanonicalPhaseTypeNHPP(3)
    model.fit(data, initialize=True, maxiter=50)

    rng = np.random.default_rng(1)
    sample = eic_sample(model, bsample=3, initialize=True, rng=rng)
    assert sample.shape == (3,)
    assert np.all(np.isfinite(sample))



"""Log-Gumbel (Frechet/Weibull as min variant) distribution helpers."""
from __future__ import annotations

import numpy as np
from scipy.stats import gumbel_r, gumbel_l

__all__ = ["dlgumbel", "plgumbel", "qlgumbel", "rlgumbel"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _dist(min: bool, loclog: float, scalelog: float):
    return gumbel_l(loc=-loclog, scale=scalelog) if min else gumbel_r(loc=loclog, scale=scalelog)


def dlgumbel(x, loclog: float = 0.0, scalelog: float = 1.0, log: bool = False, min: bool = False):
    """Density of log-Gumbel distribution (max/min)."""
    x = _as_float_array(x)
    loclog = float(loclog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    dist = _dist(min, loclog, scalelog)
    pos = x > 0.0
    if log:
        out = np.full_like(x, -np.inf, dtype=float)
        logx = np.log(x[pos])
        out[pos] = dist.logpdf(logx) - logx
        return out
    out = np.zeros_like(x, dtype=float)
    logx = np.log(x[pos])
    out[pos] = dist.pdf(logx) / x[pos]
    return out


def plgumbel(q, loclog: float = 0.0, scalelog: float = 1.0, lower_tail: bool = True, log_p: bool = False, min: bool = False):
    """CDF / upper-tail probability for log-Gumbel distribution (max/min)."""
    q = _as_float_array(q)
    loclog = float(loclog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    dist = _dist(min, loclog, scalelog)
    pos = q > 0.0
    if lower_tail:
        v = np.zeros_like(q, dtype=float)
        if np.any(pos):
            v[pos] = dist.cdf(np.log(q[pos]))
        if log_p:
            return np.log(v)
        return v
    v = np.ones_like(q, dtype=float)
    if np.any(pos):
        v[pos] = dist.sf(np.log(q[pos]))
    if log_p:
        return np.log(v)
    return v


def qlgumbel(p, loclog: float = 0.0, scalelog: float = 1.0, lower_tail: bool = True, log_p: bool = False, min: bool = False):
    """Quantile for log-Gumbel distribution (max/min)."""
    p = _as_float_array(p)
    loclog = float(loclog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    if log_p:
        p = np.exp(p)
    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")
    dist = _dist(min, loclog, scalelog)
    if lower_tail:
        return np.exp(dist.ppf(p))
    return np.exp(dist.isf(p))


def rlgumbel(n: int, loclog: float = 0.0, scalelog: float = 1.0, min: bool = False, rng: np.random.Generator | None = None):
    """Random variates from log-Gumbel distribution (max/min)."""
    if n < 0:
        raise ValueError("n must be >= 0.")
    loclog = float(loclog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    rng = rng if rng is not None else np.random.default_rng()
    dist = _dist(min, loclog, scalelog)
    return np.exp(dist.rvs(size=int(n), random_state=rng))



from __future__ import annotations

import numpy as np

from ..dists.exp import dexp, pexp

from ..base import NHPPModel
from ..data import NHPPData
from ... import _core


class ExponentialNHPP(NHPPModel):
    name = "exp"
    df = 2

    def __init__(self, *, omega0: float = 1.0, rate0: float = 1.0):
        super().__init__(omega0=float(omega0), rate0=float(rate0))
        self._set_fitted_params(np.array([omega0, rate0], dtype=float))

    def param_names(self) -> list[str]:
        return ["omega", "rate"]

    def init_params(self, data: NHPPData) -> np.ndarray:
        return np.array([data.total, 1.0 / data.mean], dtype=float)

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        return _core.em_exp_emstep(np.asarray(params, dtype=float), data.to_core_dict())

    def omega_(self) -> float:
        return float(self.params_[0])

    def rate_(self) -> float:
        return float(self.params_[1])

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * pexp(t, rate=self.rate_(), lower_tail=True, log_p=False)

    def dmvf(self, t):
        t = np.asarray(t, dtype=float)
        t2 = np.concatenate([[0.0], t])
        return np.diff(self.mvf(t2))

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dexp(t, rate=self.rate_(), log=False)





from __future__ import annotations

import numpy as np


def nhpp_options():
    return {
        "maxiter": 10000,
        "reltol": float(np.sqrt(np.finfo(float).eps)),
        "abstol": 1.0e200,
        "stopcond": "llf",
        "trace": False,
        "printsteps": 50,
    }



from __future__ import annotations

import numpy as np
from scipy.stats import expon

__all__ = ["dexp", "pexp", "qexp", "rexp"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def dexp(x, rate: float = 1.0, log: bool = False):
    """
    Density of Exponential(rate).
    """
    x = _as_float_array(x)
    rate = float(rate)
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    scale = 1.0 / rate
    if log:
        return expon.logpdf(x, scale=scale)
    return expon.pdf(x, scale=scale)


def pexp(q, rate: float = 1.0, lower_tail: bool = True, log_p: bool = False):
    """
    CDF / upper-tail probability for Exponential(rate).
    """
    q = _as_float_array(q)
    rate = float(rate)
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    scale = 1.0 / rate
    if lower_tail:
        v = expon.cdf(q, scale=scale)
    else:
        v = expon.sf(q, scale=scale)
    if log_p:
        return np.log(v)
    return v


def qexp(p, rate: float = 1.0, lower_tail: bool = True, log_p: bool = False):
    """
    Quantile for Exponential(rate).
    """
    p = _as_float_array(p)
    rate = float(rate)
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    if log_p:
        p = np.exp(p)
    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")
    scale = 1.0 / rate
    if lower_tail:
        return expon.ppf(p, scale=scale)
    return expon.isf(p, scale=scale)


def rexp(n: int, rate: float = 1.0, rng: np.random.Generator | None = None):
    """
    Random variates from Exponential(rate).
    """
    if n < 0:
        raise ValueError("n must be >= 0.")
    rate = float(rate)
    if rate <= 0:
        raise ValueError("rate must be > 0.")
    rng = rng if rng is not None else np.random.default_rng()
    scale = 1.0 / rate
    return expon.rvs(scale=scale, size=int(n), random_state=rng)



from pysrat.nhpp.data import NHPPData


def test_nhppdata_constructors():
    d1 = NHPPData.from_intervals(time=[1, 1, 1, 1], fault=[0, 1, 0, 5])
    assert d1.len == 4
    assert d1.total == 6

    d2 = NHPPData.from_fault_times([3, 1, 7, 15, 12], te=3)
    assert d2.len == 6

    d3 = NHPPData.from_counts([0, 1, 0, 5])
    assert d3.total == 6



import numpy as np
import pytest


marlib_cf1 = pytest.importorskip("pysrat.marlib_cf1")


def _params():
    alpha = np.array([0.4, 0.3, 0.2], dtype=float)
    rate = np.array([1.5, 0.5, 1.0], dtype=float)
    return alpha, rate


def _dx():
    return np.array([0.1, 0.5, 1.0], dtype=float)


def test_cf1pdf_and_cf1cdf():
    alpha, rate = _params()
    dx = _dx()

    pdf = marlib_cf1.cf1pdf(dx, alpha, rate)
    assert pdf.shape == dx.shape
    assert np.all(np.isfinite(pdf))

    cdf = marlib_cf1.cf1cdf(dx, alpha, rate)
    assert cdf.shape == dx.shape
    assert np.all(np.isfinite(cdf))
    assert np.all((cdf >= 0.0) & (cdf <= 1.0))

    logpdf = marlib_cf1.cf1pdf(dx, alpha, rate, log=True)
    assert logpdf.shape == dx.shape
    assert np.all(np.isfinite(logpdf))

    logcdf = marlib_cf1.cf1cdf(dx, alpha, rate, log=True, lower=False)
    assert logcdf.shape == dx.shape
    assert np.all(np.isfinite(logcdf))


def test_cf1sample():
    alpha, rate = _params()
    n = 10
    out = marlib_cf1.cf1sample(n, alpha, rate, seed=1234)
    assert out.shape == (n,)
    assert np.all(np.isfinite(out))
    assert np.all(out >= 0.0)


def test_cf1reform():
    alpha, rate = _params()
    out = marlib_cf1.cf1reform(alpha, rate)
    assert set(out.keys()) == {"alpha", "rate"}

    a = np.asarray(out["alpha"], dtype=float)
    r = np.asarray(out["rate"], dtype=float)
    assert a.shape == alpha.shape
    assert r.shape == rate.shape
    assert np.all(np.diff(r) >= 0.0)


def test_cf1sojourn():
    alpha, rate = _params()
    f = np.array([1.0, 0.5, 0.25], dtype=float)
    b = np.array([0.2, 0.3, 0.4], dtype=float)
    out = marlib_cf1.cf1sojourn(alpha, rate, f, b, t=1.0)
    assert out.shape == (2 * alpha.size,)
    assert np.all(np.isfinite(out))


def test_em_cf1_emstep():
    alpha, rate = _params()
    params = {
        "omega": 10.0,
        "alpha": alpha,
        "rate": rate,
    }
    data = {
        "time": np.array([1.0, 2.0, 3.0], dtype=float),
        "fault": np.array([1, 0, 2], dtype=np.int64),
        "type": np.array([0, 1, 0], dtype=np.int64),
    }

    out = marlib_cf1.em_cf1_emstep(params, data)
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}

    param = out["param"]
    pdiff = out["pdiff"]

    assert set(param.keys()) == {"omega", "alpha", "rate"}
    assert set(pdiff.keys()) == {"omega", "alpha", "rate"}

    new_alpha = np.asarray(param["alpha"], dtype=float)
    new_rate = np.asarray(param["rate"], dtype=float)

    assert new_alpha.shape == alpha.shape
    assert new_rate.shape == rate.shape
    assert np.isfinite(out["llf"])
    assert np.isfinite(out["total"])



import pysrat
from pysrat.nhpp.data import NHPPData
from pysrat.nhpp.models.pareto2 import Pareto2NHPP


def test_model_fit_api_pareto2():
    d = NHPPData.from_counts([0, 1, 0, 5])
    model = Pareto2NHPP().fit(d, maxiter=5)
    assert model.params_.shape == (3,)
    assert hasattr(model, "aic_")


def test_plot_accepts_model_pareto2():
    d = NHPPData.from_counts([0, 1, 0, 5])
    model = Pareto2NHPP().fit(d, maxiter=5)
    ax = pysrat.plot_mvf(d, model)
    assert ax is not None



from __future__ import annotations

import numpy as np
from scipy.optimize import minimize


def optimize_params(
    pllf_fn,
    data_dict: dict,
    x0: np.ndarray,
    *,
    w0=None,
    w1=None,
    method: str = "L-BFGS-B",
    bounds=None,
    options: dict | None = None,
) -> np.ndarray:
    def obj(x):
        loc, scale = float(x[0]), float(x[1])
        if scale <= 0:
            return np.inf
        params = np.asarray([loc, scale], dtype=float)
        if w0 is None:
            return -float(pllf_fn(params, data_dict, float(w1)))
        return -float(pllf_fn(params, data_dict, float(w0), float(w1)))

    if method.upper() == "L-BFGS-B" and bounds is None:
        bounds = [(None, None), (1.0e-12, None)]
    res = minimize(
        obj,
        x0=np.asarray(x0, dtype=float),
        method=method,
        bounds=bounds,
        options=options,
    )
    if not res.success or not np.all(np.isfinite(res.x)):
        return np.asarray(x0, dtype=float)
    return np.asarray(res.x, dtype=float)



import numpy as np
from pysrat.nhpp.data import NHPPData
from pysrat.nhpp.dists.tnorm import dtnorm, ptnorm, qtnorm, rtnorm
from pysrat.nhpp.models.tnorm import TruncatedNormalNHPP


def test_tnorm_distribution_basic():
    x = np.array([0.0, 0.5, 1.0])
    d = dtnorm(x, mean=0.0, sd=1.0)
    p = ptnorm(x, mean=0.0, sd=1.0)
    q = qtnorm([0.1, 0.5, 0.9], mean=0.0, sd=1.0)
    r = rtnorm(5, mean=0.0, sd=1.0)

    assert d.shape == x.shape
    assert p.shape == x.shape
    assert q.shape == (3,)
    assert r.shape == (5,)
    assert np.all(r >= 0.0)


def test_tnorm_model_fit_smoke():
    data = NHPPData.from_intervals(
        time=[1.0, 2.0, 1.5],
        fault=[0.0, 2.0, 1.0],
        type=[0, 1, 0],
    )
    model = TruncatedNormalNHPP().fit(data, maxiter=5)
    assert model.params_.shape == (3,)
    assert hasattr(model, "aic_")



from __future__ import annotations

import numpy as np

from ..base import NHPPModel
from ..data import NHPPData
from ..dists.gamma import dgamma, pgamma
from ... import _core


class GammaNHPP(NHPPModel):
    name = "gamma"
    df = 3

    def __init__(self, *, omega0: float = 1.0, shape0: float = 1.0, rate0: float = 1.0):
        super().__init__(omega0=float(omega0), shape0=float(shape0), rate0=float(rate0))
        self._set_fitted_params(np.array([omega0, shape0, rate0], dtype=float))

    def param_names(self) -> list[str]:
        return ["omega", "shape", "rate"]

    def init_params(self, data: NHPPData) -> np.ndarray:
        return np.array([data.total, 1.0, 1.0 / data.mean], dtype=float)

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        divide = int(kwargs.get("divide", 15))
        eps = float(kwargs.get("eps", 1.0e-10))
        return _core.em_gamma_emstep(np.asarray(params, dtype=float), data.to_core_dict(), divide, eps)

    def omega_(self) -> float:
        return float(self.params_[0])

    def shape_(self) -> float:
        return float(self.params_[1])

    def rate_(self) -> float:
        return float(self.params_[2])

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * pgamma(t, shape=self.shape_(), rate=self.rate_(), lower_tail=True)

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dgamma(t, shape=self.shape_(), rate=self.rate_(), log=False)





from __future__ import annotations

import numpy as np

from ..base import NHPPModel
from ..data import NHPPData
from ..dists.tlogis import dtlogis, ptlogis
from ... import _core
from ._utils import optimize_params


class TruncatedLogisticNHPP(NHPPModel):
    name = "tlogis"
    df = 3

    def __init__(self, *, omega0: float = 1.0, location0: float = 0.0, scale0: float = 1.0):
        super().__init__(omega0=float(omega0), location0=float(location0), scale0=float(scale0))
        self._set_fitted_params(np.array([omega0, location0, scale0], dtype=float))

    def param_names(self) -> list[str]:
        return ["omega", "location", "scale"]

    def init_params(self, data: NHPPData) -> np.ndarray:
        return np.array([1.0, 0.0, data.mean], dtype=float)

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        core = data.to_core_dict()
        eres = _core.em_tlogis_estep(np.asarray(params, dtype=float), core)
        method = kwargs.get("method", "L-BFGS-B")
        options = kwargs.get("options")
        loc_scale = optimize_params(
            _core.em_tlogis_pllf,
            core,
            np.asarray(params[1:3], dtype=float),
            w0=float(eres["w0"]),
            w1=float(eres["w1"]),
            method=method,
            options=options,
        )
        new_param = np.array([float(eres["omega"]), loc_scale[0], loc_scale[1]], dtype=float)
        return {
            "param": new_param,
            "pdiff": new_param - np.asarray(params, dtype=float),
            "llf": float(eres["llf"]),
            "total": float(eres["total"]),
        }

    def omega_(self) -> float:
        return float(self.params_[0])

    def location_(self) -> float:
        return float(self.params_[1])

    def scale_(self) -> float:
        return float(self.params_[2])

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * ptlogis(t, location=self.location_(), scale=self.scale_(), lower_tail=True)

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dtlogis(t, location=self.location_(), scale=self.scale_(), log=False)





import pysrat
from pysrat.nhpp.data import NHPPData
from pysrat.nhpp.models.exp import ExponentialNHPP
from pysrat.nhpp.fit import compare

def test_model_fit_api_exponential():
    d = NHPPData.from_counts([0, 1, 0, 5])
    model = ExponentialNHPP().fit(d, maxiter=5)
    assert model.params_.shape == (2,)
    assert hasattr(model, "aic_")


def test_plot_accepts_model_exponential():
    d = NHPPData.from_counts([0, 1, 0, 5])
    model = ExponentialNHPP().fit(d, maxiter=5)
    ax = pysrat.plot_mvf(d, model)
    assert ax is not None


def test_compare_models_exponential():
    d = NHPPData.from_counts([0, 1, 0, 5])
    fitted, best = compare([ExponentialNHPP()], d)
    assert best is fitted[0]



from pysrat.nhpp.data import NHPPData
from pysrat.nhpp.models.txvmax import TruncatedExtremeValueMaxNHPP


def test_txvmax_model_fit_smoke():
    data = NHPPData.from_intervals(
        time=[1.0, 2.0, 1.5, 0.5],
        fault=[0.0, 1.0, 0.0, 2.0],
        type=[0, 1, 0, 0],
    )
    model = TruncatedExtremeValueMaxNHPP().fit(data, maxiter=5)
    assert model.params_.shape == (3,)
    assert hasattr(model, "aic_")



import csv
import importlib.resources as resources

import pandas as pd

from pysrat.nhpp.data import NHPPData
from pysrat.nhpp.models.exp import ExponentialNHPP


def test_musa_dataset_available():
    data_path = resources.files("pysrat").joinpath("nhpp/datasets/musa/ss1a.csv")
    assert data_path.is_file()

    with data_path.open("r", newline="") as f:
        reader = csv.reader(f)
        header = next(reader)
        first_row = next(reader)

    assert header == ["time", "fault", "indicator"]
    assert len(first_row) == 3


def test_musa_ss1a_exp_fit():
    data_path = resources.files("pysrat").joinpath("nhpp/datasets/musa/ss1a.csv")
    df = pd.read_csv(data_path)
    data = NHPPData.from_intervals(
        time=df["time"].to_numpy(dtype=float),
        fault=df["fault"].to_numpy(dtype=float),
        type=df["indicator"].to_numpy(dtype=int),
    )
    model = ExponentialNHPP().fit(data, maxiter=5)
    assert model.params_.shape == (2,)
    assert hasattr(model, "aic_")



from __future__ import annotations

from dataclasses import dataclass
import numpy as np


@dataclass(frozen=True)
class NHPPData:
	time: np.ndarray
	fault: np.ndarray
	type: np.ndarray
	total: int
	mean: float
	max: float
	kind: str

	@property
	def len(self) -> int:
		return int(np.asarray(self.time).shape[0])

	def to_core_dict(self) -> dict:
		return {
			"len": self.len,
			"time": np.asarray(self.time, dtype=float),
			"fault": np.asarray(self.fault, dtype=float),
			"type": np.asarray(self.type, dtype=np.int64),
		}

	@classmethod
	def from_intervals(cls, *, time=None, fault=None, type=None, te=None) -> "NHPPData":
		time = None if time is None else np.asarray(time, dtype=float)
		fault = None if fault is None else np.asarray(fault, dtype=int)
		type = None if type is None else np.asarray(type, dtype=int)
		te = None if te is None else float(te)

		if time is None:
			if fault is None:
				raise ValueError("Invalid data: Either time or fault is required.")
			if type is None:
				type = np.zeros_like(fault, dtype=int)
			time = np.ones_like(fault, dtype=float)
		else:
			if fault is None:
				if type is None:
					if te is None:
						raise ValueError("Invalid data: Either type or te is required when fault is missing.")
					type = np.ones_like(time, dtype=int)
					time = np.concatenate([time, np.array([te], dtype=float)])
					type = np.concatenate([type, np.array([0], dtype=int)])
					fault = np.zeros_like(time, dtype=int)
				else:
					fault = np.zeros_like(time, dtype=int)
			else:
				if type is None:
					type = np.zeros_like(time, dtype=int)

		if time.shape[0] != fault.shape[0] or time.shape[0] != type.shape[0]:
			raise ValueError("Invalid data")
		if np.all((fault == 0) & (type == 0)):
			raise ValueError("Invalid data: no fault.")
		if np.any((time == 0) & (fault != 0) & (type != 0)):
			raise ValueError("Invalid data: zero time exits.")

		tmp = fault + type
		total = int(tmp.sum())
		ct = np.cumsum(time)
		mean = float((ct * tmp).sum() / total)
		maxv = float(ct[tmp >= 1].max())

		return cls(
			time=np.asarray(time, dtype=float),
			fault=np.asarray(fault, dtype=int),
			type=np.asarray(type, dtype=int),
			total=total,
			mean=mean,
			max=maxv,
			kind="intervals",
		)

	@classmethod
	def from_counts(cls, fault, *, type=None) -> "NHPPData":
		data = cls.from_intervals(time=None, fault=fault, type=type, te=None)
		return cls(
			time=data.time,
			fault=data.fault,
			type=data.type,
			total=data.total,
			mean=data.mean,
			max=data.max,
			kind="counts",
		)

	@classmethod
	def from_fault_times(cls, times, *, te) -> "NHPPData":
		data = cls.from_intervals(time=times, fault=None, type=None, te=te)
		return cls(
			time=data.time,
			fault=data.fault,
			type=data.type,
			total=data.total,
			mean=data.mean,
			max=data.max,
			kind="fault_times",
		)

	def __repr__(self) -> str:
		cols = np.vstack([self.time, self.fault, self.type]).T
		head = "NHPPData(time, fault, type)\n"
		body = "\n".join([f"{r[0]:g}\t{int(r[1])}\t{int(r[2])}" for r in cols[:50]])
		if cols.shape[0] > 50:
			body += "\n..."
		meta = f"\nlen={self.len}, total={self.total}, mean={self.mean:g}, max={self.max:g}"
		return head + "time\tfault\ttype\n" + body + meta


__all__ = ["NHPPData"]



"""Truncated Gumbel (max/min) distribution at 0."""
from __future__ import annotations

import numpy as np
from scipy.stats import gumbel_r, gumbel_l

__all__ = ["dtgumbel", "ptgumbel", "qtgumbel", "rtgumbel"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _dist(min: bool, loc: float, scale: float):
    return gumbel_l(loc=-loc, scale=scale) if min else gumbel_r(loc=loc, scale=scale)


def dtgumbel(x, loc: float = 0.0, scale: float = 1.0, log: bool = False, min: bool = False):
    """Density of Gumbel distribution truncated to [0, +inf) (max/min)."""
    x = _as_float_array(x)
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    dist = _dist(min, loc, scale)
    denom = dist.sf(0.0)
    if log:
        out = dist.logpdf(x) - np.log(denom)
        return np.where(x < 0.0, -np.inf, out)
    out = dist.pdf(x) / denom
    return np.where(x < 0.0, 0.0, out)


def ptgumbel(q, loc: float = 0.0, scale: float = 1.0, lower_tail: bool = True, log_p: bool = False, min: bool = False):
    """CDF / upper-tail probability for truncated Gumbel distribution (max/min)."""
    q = _as_float_array(q)
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    dist = _dist(min, loc, scale)
    denom = dist.sf(0.0)
    upper = dist.sf(q) / denom
    if lower_tail:
        v = 1.0 - upper
        v = np.where(q < 0.0, 0.0, v)
    else:
        v = upper
        v = np.where(q < 0.0, 1.0, v)
    if log_p:
        return np.log(v)
    return v


def qtgumbel(p, loc: float = 0.0, scale: float = 1.0, lower_tail: bool = True, log_p: bool = False, min: bool = False):
    """Quantile for truncated Gumbel distribution (max/min)."""
    p = _as_float_array(p)
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    if log_p:
        p = np.exp(p)
    if not lower_tail:
        p = 1.0 - p
    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")
    dist = _dist(min, loc, scale)
    pdash = (1.0 - p) * dist.sf(0.0)
    out = dist.isf(pdash)
    return np.maximum(out, 0.0)


def rtgumbel(n: int, loc: float = 0.0, scale: float = 1.0, min: bool = False, rng: np.random.Generator | None = None):
    """Random variates from truncated Gumbel distribution (max/min)."""
    if n < 0:
        raise ValueError("n must be >= 0.")
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    rng = rng if rng is not None else np.random.default_rng()
    u = rng.uniform(0.0, 1.0, size=int(n))
    return qtgumbel(u, loc=loc, scale=scale, lower_tail=True, log_p=False, min=min)



from __future__ import annotations

from dataclasses import dataclass
import copy
import numpy as np
from scipy.optimize import minimize_scalar
from scipy.stats import t as _t_dist

from .base import NHPPModel
from .data import NHPPData

__all__ = [
    "FitResult",
    "emfit",
    "bs_time",
    "bs_group",
    "eic_group",
    "eic_time",
    "eic_sample",
    "eic",
]


@dataclass
class FitResult:
    srm: NHPPModel
    llf: float
    data: NHPPData

    @classmethod
    def from_model(cls, model: NHPPModel, data: NHPPData | None = None) -> "FitResult":
        if data is None:
            if model.data_ is None:
                raise ValueError("data must be provided when model has no fitted data")
            data = model.data_
        if model.llf_ is None:
            raise ValueError("model.llf_ is not set; call fit() first")
        return cls(srm=model, llf=float(model.llf_), data=data)


def _clone_model(model: NHPPModel) -> NHPPModel:
    params = copy.deepcopy(model.get_params())
    return model.__class__(**params)


def emfit(model: NHPPModel, data: NHPPData, *, initialize: bool = True, **kwargs) -> FitResult:
    cloned = _clone_model(model)
    cloned.fit(data, initialize=initialize, **kwargs)
    return FitResult(srm=cloned, llf=float(cloned.llf_), data=data)


def _as_fitresult(obj: FitResult | NHPPModel, data: NHPPData | None = None) -> FitResult:
    if isinstance(obj, FitResult):
        return obj
    if isinstance(obj, NHPPModel):
        return FitResult.from_model(obj, data=data)
    raise TypeError("obj must be FitResult or NHPPModel")


def _intensity_scalar(model: NHPPModel, t: float) -> float:
    return float(np.asarray(model.intensity(np.array([t], dtype=float)))[0])


def bs_time(srm: NHPPModel, data: NHPPData, maximum: float, *, rng: np.random.Generator | None = None) -> NHPPData:
    if maximum <= 0:
        raise ValueError("maximum must be > 0")
    rng = rng if rng is not None else np.random.default_rng()

    cte = float(np.sum(data.time))

    n = rng.poisson(maximum * cte)
    if n == 0:
        raise ValueError("bootstrap produced no events; increase maximum or retry")

    candidates = np.sort(rng.uniform(0.0, cte, size=int(n)))
    keep = srm.intensity(candidates) / maximum > rng.uniform(0.0, 1.0, size=int(n))
    y = candidates[keep]
    if y.size == 0:
        raise ValueError("bootstrap produced no events; increase maximum or retry")

    time = np.diff(np.concatenate([[0.0], y]))
    te_res = float(cte - np.sum(time))
    return NHPPData.from_intervals(time=time, te=te_res)


def bs_group(srm: NHPPModel, data: NHPPData, *, rng: np.random.Generator | None = None) -> NHPPData:
    rng = rng if rng is not None else np.random.default_rng()

    ctime = np.concatenate([[0.0], np.cumsum(np.asarray(data.time, dtype=float))])
    lam = np.diff(srm.mvf(ctime))
    lam = np.clip(lam, 0.0, None)
    fault = rng.poisson(lam)
    return NHPPData.from_counts(fault=fault)


def _ensure_llf(model: NHPPModel):
    if not hasattr(model, "llf"):
        raise NotImplementedError("model.llf is required for EIC bootstrap")


def _faultdata_kind(data: NHPPData) -> str:
    return data.kind


def eic_group(obj: FitResult | NHPPModel, bsample: int = 100, *, initialize: bool = False,
              rng: np.random.Generator | None = None, data: NHPPData | None = None) -> np.ndarray:
    obj = _as_fitresult(obj, data=data)
    _ensure_llf(obj.srm)

    rng = rng if rng is not None else np.random.default_rng()

    b3 = float(obj.llf)
    b1 = np.zeros(int(bsample), dtype=float)
    b2 = np.zeros(int(bsample), dtype=float)
    b4 = np.zeros(int(bsample), dtype=float)

    data = obj.data
    ctime = np.concatenate([[0.0], np.cumsum(np.asarray(data.time, dtype=float))])
    lam = np.diff(obj.srm.mvf(ctime))

    for b in range(int(bsample)):
        fault = rng.poisson(lam)
        sample = NHPPData.from_counts(fault=fault)
        b2[b] = float(obj.srm.llf(sample))
        obj_bs = emfit(obj.srm, sample, initialize=initialize)
        b1[b] = float(obj_bs.llf)
        b4[b] = float(obj_bs.srm.llf(obj.data))

    return b1 - b2 + b3 - b4


def eic_time(obj: FitResult | NHPPModel, bsample: int = 100, *, initialize: bool = False,
             rng: np.random.Generator | None = None, data: NHPPData | None = None) -> np.ndarray:
    obj = _as_fitresult(obj, data=data)
    _ensure_llf(obj.srm)

    rng = rng if rng is not None else np.random.default_rng()

    b3 = float(obj.llf)
    b1 = np.zeros(int(bsample), dtype=float)
    b2 = np.zeros(int(bsample), dtype=float)
    b4 = np.zeros(int(bsample), dtype=float)

    data = obj.data
    cte = float(np.sum(data.time))

    xs = np.linspace(0.0, cte, 2048)
    vals = np.asarray(obj.srm.intensity(xs), dtype=float)
    maximum = float(np.max(vals)) * 1.05

    for b in range(int(bsample)):
        sample = bs_time(obj.srm, data, maximum, rng=rng)
        b2[b] = float(obj.srm.llf(sample))
        obj_bs = emfit(obj.srm, sample, initialize=initialize)
        b1[b] = float(obj_bs.llf)
        b4[b] = float(obj_bs.srm.llf(obj.data))

    return b1 - b2 + b3 - b4


def eic_sample(obj: FitResult | NHPPModel, bsample: int = 100, *, initialize: bool = False,
               rng: np.random.Generator | None = None, data: NHPPData | None = None) -> np.ndarray:
    obj = _as_fitresult(obj, data=data)
    kind = _faultdata_kind(obj.data)
    if kind == "fault_times":
        bias_sample = eic_time(obj, bsample, initialize=initialize, rng=rng)
    elif kind == "counts":
        bias_sample = eic_group(obj, bsample, initialize=initialize, rng=rng)
    else:
        raise ValueError("EIC supports only data from from_counts or from_fault_times.")
    return -2.0 * (float(obj.llf) - bias_sample)


def eic(obj: FitResult | NHPPModel, bsample: int = 100, alpha: float = 0.95, *, initialize: bool = False,
    rng: np.random.Generator | None = None, data: NHPPData | None = None) -> dict:
    obj = _as_fitresult(obj, data=data)
    if not (0.0 < alpha < 1.0):
        raise ValueError("alpha must be in (0,1)")
    kind = _faultdata_kind(obj.data)
    if kind == "fault_times":
        bias_sample = eic_time(obj, bsample, initialize=initialize, rng=rng)
    elif kind == "counts":
        bias_sample = eic_group(obj, bsample, initialize=initialize, rng=rng)
    else:
        raise ValueError("EIC supports only data from from_counts or from_fault_times.")

    x = float(np.mean(bias_sample))
    s = float(np.std(bias_sample, ddof=1))
    df = int(bsample) - 1
    talpha = float(_t_dist.ppf(1.0 - (1.0 - alpha) / 2.0, df=df))
    x_interval = x + np.array([-talpha, talpha], dtype=float) * s / np.sqrt(float(bsample))

    return {
        "bias": x,
        "bias.lower": float(x_interval[0]),
        "bias.upper": float(x_interval[1]),
        "eic": -2.0 * (float(obj.llf) - x),
        "eic.lower": -2.0 * (float(obj.llf) - float(x_interval[0])),
        "eic.upper": -2.0 * (float(obj.llf) - float(x_interval[1])),
    }



import numpy as np

from pysrat import _core


def _sample_data():
    return {
        "len": 3,
        "time": np.array([1.0, 1.0, 1.0], dtype=float),
        "fault": np.array([1.0, 0.0, 2.0], dtype=float),
        "type": np.array([0, 1, 0], dtype=np.int64),
    }


def test_em_exp_emstep():
    params = np.array([10.0, 0.2], dtype=float)
    out = _core.em_exp_emstep(params, _sample_data())
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (2,)


def test_em_tnorm_emstep():
    params = np.array([10.0, 0.0, 1.0], dtype=float)
    out = _core.em_tnorm_emstep(params, _sample_data())
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (3,)


def test_em_pareto_emstep():
    params = np.array([10.0, 1.5, 2.0], dtype=float)
    out = _core.em_pareto_emstep(params, _sample_data())
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (3,)


def test_em_lnorm_emstep():
    params = np.array([10.0, 0.0, 1.0], dtype=float)
    out = _core.em_lnorm_emstep(params, _sample_data())
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (3,)


def test_em_tlogis_emstep_mo():
    params = np.array([10.0, 0.0, 1.0], dtype=float)
    out = _core.em_tlogis_emstep_mo(params, _sample_data())
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (3,)


def test_em_tlogis_estep():
    params = np.array([10.0, 0.0, 1.0], dtype=float)
    out = _core.em_tlogis_estep(params, _sample_data())
    assert set(out.keys()) == {"llf", "omega", "w0", "w1", "total"}


def test_em_tlogis_pllf():
    params = np.array([0.0, 1.0], dtype=float)
    out = _core.em_tlogis_pllf(params, _sample_data(), 1.0, 1.0)
    assert np.isfinite(out)


def test_em_llogis_emstep():
    params = np.array([10.0, 0.0, 1.0], dtype=float)
    out = _core.em_llogis_emstep(params, _sample_data())
    assert set(out.keys()) == {"param", "pdiff", "llf", "total", "residual"}
    assert out["param"].shape == (3,)


def test_em_llogis_estep():
    params = np.array([10.0, 0.0, 1.0], dtype=float)
    out = _core.em_llogis_estep(params, _sample_data())
    assert set(out.keys()) == {"llf", "omega", "w1", "total"}


def test_em_llogis_pllf():
    params = np.array([0.0, 1.0], dtype=float)
    out = _core.em_llogis_pllf(params, _sample_data(), 1.0)
    assert np.isfinite(out)


def test_em_gamma_emstep():
    params = np.array([10.0, 1.5, 2.0], dtype=float)
    out = _core.em_gamma_emstep(params, _sample_data(), divide=5, eps=1.0e-8)
    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (3,)



from __future__ import annotations

import numpy as np

from ..base import NHPPModel
from ..data import NHPPData
from ..dists.lnorm import dlnorm, plnorm
from ... import _core


class LogNormalNHPP(NHPPModel):
    name = "lnorm"
    df = 3

    def __init__(self, *, omega0: float = 1.0, meanlog0: float = 0.0, sdlog0: float = 1.0):
        super().__init__(omega0=float(omega0), meanlog0=float(meanlog0), sdlog0=float(sdlog0))
        self._set_fitted_params(np.array([omega0, meanlog0, sdlog0], dtype=float))

    def param_names(self) -> list[str]:
        return ["omega", "meanlog", "sdlog"]

    def init_params(self, data: NHPPData) -> np.ndarray:
        return np.array([1.0, 1.0, max(np.log(data.mean), 1.0)], dtype=float)

    def em_step(self, params: np.ndarray, data: NHPPData, **kwargs) -> dict:
        return _core.em_lnorm_emstep(np.asarray(params, dtype=float), data.to_core_dict())

    def omega_(self) -> float:
        return float(self.params_[0])

    def meanlog_(self) -> float:
        return float(self.params_[1])

    def sdlog_(self) -> float:
        return float(self.params_[2])

    def mvf(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * plnorm(t, meanlog=self.meanlog_(), sdlog=self.sdlog_(), lower_tail=True)

    def intensity(self, t):
        t = np.asarray(t, dtype=float)
        return self.omega_() * dlnorm(t, meanlog=self.meanlog_(), sdlog=self.sdlog_(), log=False)





from ._version import __version__
from .data import NHPPData  # noqa: F401
from .base import NHPPModel  # noqa: F401
from .models.exp import ExponentialNHPP  # noqa: F401
from .models.pareto2 import Pareto2NHPP  # noqa: F401
from .models.gamma import GammaNHPP  # noqa: F401
from .models.lnorm import LogNormalNHPP  # noqa: F401
from .models.tlogis import TruncatedLogisticNHPP  # noqa: F401
from .models.llogis import LogLogisticNHPP  # noqa: F401
from .models.tnorm import TruncatedNormalNHPP  # noqa: F401
from .models.cf1 import CanonicalPhaseTypeNHPP  # noqa: F401
from .models.txvmax import TruncatedExtremeValueMaxNHPP  # noqa: F401
from .models.lxvmax import LogExtremeValueMaxNHPP  # noqa: F401
from .models.txvmin import TruncatedExtremeValueMinNHPP  # noqa: F401
from .models.lxvmin import LogExtremeValueMinNHPP  # noqa: F401
from .fit import compare  # noqa: F401
from .bootstrap import (  # noqa: F401
	FitResult,
	emfit,
	bs_time,
	bs_group,
	eic_group,
	eic_time,
	eic_sample,
	eic,
)
from .options import nhpp_options  # noqa: F401
from .plot import plot_mvf, plot_dmvf, plot_rate  # noqa: F401



import numpy as np

from pysrat.nhpp.dists import (
    dexp,
    pexp,
    qexp,
    rexp,
    dtnorm,
    ptnorm,
    qtnorm,
    rtnorm,
    dpareto2,
    ppareto2,
    qpareto2,
    rpareto2,
    dlnorm,
    plnorm,
    qlnorm,
    rlnorm,
    dtlogis,
    ptlogis,
    qtlogis,
    rtlogis,
    dllogis,
    pllogis,
    qllogis,
    rllogis,
    dgumbel,
    pgumbel,
    qgumbel,
    rgumbel,
    dtgumbel,
    ptgumbel,
    qtgumbel,
    rtgumbel,
    dlgumbel,
    plgumbel,
    qlgumbel,
    rlgumbel,
    dcf1,
    pcf1,
    rcf1,
)


def test_exp_distribution():
    x = np.array([0.1, 0.5, 1.0])
    d = dexp(x, rate=2.0)
    assert d.shape == x.shape
    assert np.all(d >= 0)

    p = pexp(x, rate=2.0)
    assert np.all((p >= 0) & (p <= 1))
    p_upper = pexp(x, rate=2.0, lower_tail=False)
    assert np.allclose(p + p_upper, 1.0, atol=1e-12)

    q = qexp(p, rate=2.0)
    assert np.allclose(q, x, atol=1e-10)

    dlog = dexp(x, rate=2.0, log=True)
    assert np.allclose(dlog, np.log(d))

    r = rexp(5, rate=2.0, rng=np.random.default_rng(0))
    assert r.shape == (5,)


def test_tnorm_distribution():
    x = np.array([-1.0, 0.0, 1.0])
    d = dtnorm(x, mean=0.0, sd=1.0)
    assert d[0] == 0.0
    assert np.all(d[1:] >= 0)

    p = ptnorm(x, mean=0.0, sd=1.0)
    assert p[0] == 0.0
    p_upper = ptnorm(x, mean=0.0, sd=1.0, lower_tail=False)
    assert p_upper[0] == 1.0

    q = qtnorm(np.array([0.0, 0.5, 1.0]), mean=0.0, sd=1.0)
    assert np.all(q >= 0.0)

    r = rtnorm(4, mean=0.0, sd=1.0, rng=np.random.default_rng(0))
    assert r.shape == (4,)


def test_pareto2_distribution():
    x = np.array([-1.0, 0.0, 1.0])
    d = dpareto2(x, shape=2.0, scale=1.0)
    assert d[0] == 0.0
    p = ppareto2(x, shape=2.0, scale=1.0)
    assert p[0] == 0.0
    p_upper = ppareto2(x, shape=2.0, scale=1.0, lower_tail=False)
    assert p_upper[0] == 1.0

    q = qpareto2(np.array([0.0, 0.5, 1.0]), shape=2.0, scale=1.0)
    assert np.all(q >= 0.0)

    r = rpareto2(3, shape=2.0, scale=1.0, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_lognorm_distribution():
    x = np.array([-1.0, 0.1, 1.0])
    d = dlnorm(x, meanlog=0.0, sdlog=1.0)
    assert d[0] == 0.0
    p = plnorm(x, meanlog=0.0, sdlog=1.0)
    assert p[0] == 0.0
    p_upper = plnorm(x, meanlog=0.0, sdlog=1.0, lower_tail=False)
    assert p_upper[0] == 1.0

    q = qlnorm(np.array([0.0, 0.5, 1.0]), meanlog=0.0, sdlog=1.0)
    assert np.all(q >= 0.0)

    r = rlnorm(3, meanlog=0.0, sdlog=1.0, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_tlogis_distribution():
    x = np.array([-1.0, 0.0, 1.0])
    d = dtlogis(x, location=0.0, scale=1.0)
    assert d[0] == 0.0
    p = ptlogis(x, location=0.0, scale=1.0)
    assert p[0] == 0.0
    p_upper = ptlogis(x, location=0.0, scale=1.0, lower_tail=False)
    assert p_upper[0] == 1.0

    q = qtlogis(np.array([0.0, 0.5, 1.0]), location=0.0, scale=1.0)
    assert np.all(q >= 0.0)

    r = rtlogis(3, location=0.0, scale=1.0, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_llogis_distribution():
    x = np.array([-1.0, 0.1, 1.0])
    d = dllogis(x, locationlog=0.0, scalelog=1.0)
    assert d[0] == 0.0
    p = pllogis(x, locationlog=0.0, scalelog=1.0)
    assert p[0] == 0.0
    p_upper = pllogis(x, locationlog=0.0, scalelog=1.0, lower_tail=False)
    assert p_upper[0] == 1.0

    q = qllogis(np.array([0.0, 0.5, 1.0]), locationlog=0.0, scalelog=1.0)
    assert np.all(q >= 0.0)

    r = rllogis(3, locationlog=0.0, scalelog=1.0, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_gumbel_distribution_min_max():
    x = np.array([-1.0, 0.0, 1.0])
    d_max = dgumbel(x, loc=0.0, scale=1.0, min=False)
    d_min = dgumbel(x, loc=0.0, scale=1.0, min=True)
    assert d_max.shape == x.shape and d_min.shape == x.shape

    p_max = pgumbel(x, loc=0.0, scale=1.0, lower=True, min=False)
    p_min = pgumbel(x, loc=0.0, scale=1.0, lower=True, min=True)
    assert np.all((p_max >= 0) & (p_max <= 1))
    assert np.all((p_min >= 0) & (p_min <= 1))

    q_max = qgumbel(np.array([0.1, 0.5, 0.9]), loc=0.0, scale=1.0, min=False)
    q_min = qgumbel(np.array([0.1, 0.5, 0.9]), loc=0.0, scale=1.0, min=True)
    assert q_max.shape == (3,) and q_min.shape == (3,)

    r = rgumbel(3, loc=0.0, scale=1.0, min=False, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_tgumbel_distribution_min_max():
    x = np.array([-1.0, 0.0, 1.0])
    d = dtgumbel(x, loc=0.0, scale=1.0, min=False)
    assert d[0] == 0.0
    p = ptgumbel(x, loc=0.0, scale=1.0, min=False)
    assert p[0] == 0.0
    p_upper = ptgumbel(x, loc=0.0, scale=1.0, lower_tail=False, min=False)
    assert p_upper[0] == 1.0

    q = qtgumbel(np.array([0.0, 0.5, 1.0]), loc=0.0, scale=1.0, min=True)
    assert np.all(q >= 0.0)

    r = rtgumbel(3, loc=0.0, scale=1.0, min=True, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_lgumbel_distribution_min_max():
    x = np.array([-1.0, 0.1, 1.0])
    d = dlgumbel(x, loclog=0.0, scalelog=1.0, min=False)
    assert d[0] == 0.0
    p = plgumbel(x, loclog=0.0, scalelog=1.0, min=True)
    assert p[0] == 0.0
    p_upper = plgumbel(x, loclog=0.0, scalelog=1.0, lower_tail=False, min=True)
    assert p_upper[0] == 1.0

    q = qlgumbel(np.array([0.1, 0.5, 0.9]), loclog=0.0, scalelog=1.0, min=False)
    assert np.all(q >= 0.0)

    r = rlgumbel(3, loclog=0.0, scalelog=1.0, min=False, rng=np.random.default_rng(0))
    assert r.shape == (3,)


def test_cf1_distribution():
    x = np.array([0.1, 0.5, 1.0])
    alpha = np.array([0.4, 0.3, 0.2])
    rate = np.array([0.5, 1.0, 1.5])

    d = dcf1(x, alpha=alpha, rate=rate)
    assert d.shape == x.shape
    assert np.all(np.isfinite(d))
    assert np.all(d >= 0.0)

    p = pcf1(x, alpha=alpha, rate=rate)
    assert p.shape == x.shape
    assert np.all(np.isfinite(p))
    assert np.all((p >= 0.0) & (p <= 1.0))

    p_upper = pcf1(x, alpha=alpha, rate=rate, lower_tail=False)
    assert np.allclose(p + p_upper, 1.0, atol=1e-10)

    r = rcf1(5, alpha=alpha, rate=rate, scramble=False)
    assert r.shape == (5,)
    assert np.all(np.isfinite(r))



"""Log-logistic distribution helpers."""
from __future__ import annotations

import numpy as np
from scipy.stats import logistic

__all__ = ["dllogis", "pllogis", "qllogis", "rllogis"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _dist(locationlog: float, scalelog: float):
    return logistic(loc=locationlog, scale=scalelog)


def dllogis(x, locationlog: float = 0.0, scalelog: float = 1.0, log: bool = False):
    """Density of log-logistic distribution."""
    x = _as_float_array(x)
    locationlog = float(locationlog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    dist = _dist(locationlog, scalelog)
    logx = np.full_like(x, -np.inf, dtype=float)
    pos = x > 0.0
    logx[pos] = np.log(x[pos])
    if log:
        out = dist.logpdf(logx) - logx
        return np.where(x <= 0.0, -np.inf, out)
    out = dist.pdf(logx) / x
    return np.where(x <= 0.0, 0.0, out)


def pllogis(q, locationlog: float = 0.0, scalelog: float = 1.0, lower_tail: bool = True, log_p: bool = False):
    """CDF / upper-tail probability for log-logistic distribution."""
    q = _as_float_array(q)
    locationlog = float(locationlog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    dist = _dist(locationlog, scalelog)
    logq = np.full_like(q, -np.inf, dtype=float)
    pos = q > 0.0
    logq[pos] = np.log(q[pos])
    if lower_tail:
        v = dist.cdf(logq)
        v = np.where(q <= 0.0, 0.0, v)
        if log_p:
            return np.log(v)
        return v
    v = dist.sf(logq)
    v = np.where(q <= 0.0, 1.0, v)
    if log_p:
        return np.log(v)
    return v


def qllogis(p, locationlog: float = 0.0, scalelog: float = 1.0, lower_tail: bool = True, log_p: bool = False):
    """Quantile for log-logistic distribution."""
    p = _as_float_array(p)
    locationlog = float(locationlog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    if log_p:
        p = np.exp(p)
    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")
    dist = _dist(locationlog, scalelog)
    if lower_tail:
        return np.exp(dist.ppf(p))
    return np.exp(dist.isf(p))


def rllogis(n: int, locationlog: float = 0.0, scalelog: float = 1.0, rng: np.random.Generator | None = None):
    """Random variates from log-logistic distribution."""
    if n < 0:
        raise ValueError("n must be >= 0.")
    locationlog = float(locationlog)
    scalelog = float(scalelog)
    if scalelog <= 0:
        raise ValueError("scalelog must be > 0.")
    rng = rng if rng is not None else np.random.default_rng()
    dist = _dist(locationlog, scalelog)
    return np.exp(dist.rvs(size=int(n), random_state=rng))



import numpy as np
import pytest

from pysrat.nhpp.data import NHPPData
from pysrat.nhpp.models.cf1 import CanonicalPhaseTypeNHPP

pytest.importorskip("pysrat.marlib_cf1")


def _sample_data():
    return NHPPData.from_intervals(
        time=np.array([1.0, 2.0, 1.5], dtype=float),
        fault=np.array([1, 0, 2], dtype=int),
        type=np.array([0, 1, 0], dtype=int),
    )


def test_cf1_model_em_step_and_llf():
    data = _sample_data()
    model = CanonicalPhaseTypeNHPP(3)

    params = model.init_params(data)
    out = model.em_step(params, data)

    assert set(out.keys()) == {"param", "pdiff", "llf", "total"}
    assert out["param"].shape == (1 + 2 * model.n,)
    assert np.isfinite(out["llf"])

    model._set_fitted_params(out["param"])
    llf = model.llf(data)
    assert np.isfinite(llf)


def test_cf1_model_mvf_intensity_shapes():
    model = CanonicalPhaseTypeNHPP(3)
    t = np.array([0.2, 0.5, 1.0], dtype=float)
    model._set_fitted_params(model.init_params(_sample_data()))

    mvf = model.mvf(t)
    intensity = model.intensity(t)

    assert mvf.shape == t.shape
    assert intensity.shape == t.shape
    assert np.all(np.isfinite(mvf))
    assert np.all(np.isfinite(intensity))



"""Gumbel (max/min) distribution helpers."""
from __future__ import annotations

import numpy as np
from scipy.stats import gumbel_r, gumbel_l

__all__ = ["dgumbel", "pgumbel", "qgumbel", "rgumbel"]


def _as_float_array(x):
    return np.asarray(x, dtype=float)


def _dist(min: bool, loc: float, scale: float):
    return gumbel_l(loc=-loc, scale=scale) if min else gumbel_r(loc=loc, scale=scale)


def dgumbel(x, loc: float = 0.0, scale: float = 1.0, log: bool = False, min: bool = False):
    """Density of Gumbel distribution (max/min)."""
    x = _as_float_array(x)
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    dist = _dist(min, loc, scale)
    if log:
        return dist.logpdf(x)
    return dist.pdf(x)


def pgumbel(q, loc: float = 0.0, scale: float = 1.0, lower: bool = True, log: bool = False, min: bool = False):
    """CDF / upper-tail probability for Gumbel distribution (max/min)."""
    q = _as_float_array(q)
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    dist = _dist(min, loc, scale)
    if lower:
        v = dist.cdf(q)
        if log:
            return dist.logcdf(q)
        return v
    v = dist.sf(q)
    if log:
        return dist.logsf(q)
    return v


def qgumbel(p, loc: float = 0.0, scale: float = 1.0, lower: bool = True, log: bool = False, min: bool = False):
    """Quantile for Gumbel distribution (max/min)."""
    p = _as_float_array(p)
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    if log:
        p = np.exp(p)
    if np.any((p < 0.0) | (p > 1.0)):
        raise ValueError("p must be in [0,1].")
    dist = _dist(min, loc, scale)
    if lower:
        return dist.ppf(p)
    return dist.isf(p)


def rgumbel(n: int, loc: float = 0.0, scale: float = 1.0, min: bool = False, rng: np.random.Generator | None = None):
    """Random variates from Gumbel distribution (max/min)."""
    if n < 0:
        raise ValueError("n must be >= 0.")
    loc = float(loc)
    scale = float(scale)
    if scale <= 0:
        raise ValueError("scale must be > 0.")
    rng = rng if rng is not None else np.random.default_rng()
    dist = _dist(min, loc, scale)
    return dist.rvs(size=int(n), random_state=rng)
